import numpy as numpy
import math

from .Kernel import Kernel

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  *    This file is part of the BayesicFitting package.
#  *
#  *    BayesicFitting is free software: you can redistribute it and/or modify
#  *    it under the terms of the GNU Lesser General Public License as
#  *    published by the Free Software Foundation, either version 3 of
#  *    the License, or ( at your option ) any later version.
#  *
#  *    BayesicFitting is distributed in the hope that it will be useful,
#  *    but WITHOUT ANY WARRANTY; without even the implied warranty of
#  *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  *    GNU Lesser General Public License for more details.
#  *
#  *    The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester



class Sinc( Kernel ):
    """
    Sinc is an unbound Kernel function.

        f( x ) = sin( x ) / x.

    Sinc is not strictly positive.

    """
    def __init__( self ) :
        """
        Constructor.

        Using
            integral = 1.0
            fwhm = 2 * 1.8954942670 / &pi;
            range = inf
        """
        super( Sinc, self ).__init__( integral=1.0, fwhm=2 * 1.8954942670 / math.pi,
                                      range=math.inf )

    def result( self, x ):
        xp = numpy.where( x == 0, 1.0, x * math.pi )
        return numpy.where( x == 0, 1.0, numpy.sin( xp ) / xp )

    def partial( self, x ):
        xp = numpy.where( x == 0, 1.0, x * math.pi )
        return numpy.where( x == 0, 0.0,
                    math.pi * ( xp * numpy.cos( xp ) - numpy.sin( xp ) ) / ( xp * xp ) )

    def isBound( self ):
        return False

    def name( self ):
        return str( "Sinc: sin( x ) / x" )





#run with: python3 -m unittest TestModifiable

import unittest
import os
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt

#from StdTests import stdModeltest

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 Do Kester

class TestModifiable( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def test1( self ) :
        print( "  Test Splines" )

        t = numpy.linspace( 0, 100, 201, dtype=float )
        y = numpy.sin( 2 * math.pi * numpy.exp( t / 150 ) )

        knots =[0, 25, 50, 75, 100]

        mdl = BasicSplinesModel( knots=knots )
        mdl.setLimits( lowLimits=[-2.0], highLimits=[+2.0] )

        engs = ["galilean", "chord"]

        ns = NestedSampler( t, mdl, y, seed=1234, engines=engs )
        ns.distribution.setLimits( [0.01,100] )
        ns.verbose = 2

        evid = ns.sample( plot=self.doplot )

    def test2( self ) :
        print( "  Test Modifiable Splines" )

#        t = numpy.linspace( 0, 100, 201, dtype=float )
        t = numpy.linspace( 0, 100, 41, dtype=float )
        y = numpy.sin( 2 * math.pi * numpy.exp( t / 70 ) )

        knots =[0, 10, 25, 40, 50, 60, 75, 100]

        mdl = SplinesDynamicModel( knots=knots, dynamic=False )
        mdl.setLimits( lowLimits=[-2.0], highLimits=[+2.0] )

        engs = ["galilean", "struct"]

        ns = NestedSampler( t, mdl, y, seed=31234, engines=engs )
        ns.distribution.setLimits( [0.01,100] )
        ns.verbose = 2
#        ns.engines[1].slow = 5

        evid = ns.sample( plot=self.doplot )

        if not self.doplot : return

        cc = ['k,', 'b,', 'r,', 'g,', 'c,', 'm,']
        sl = ns.samples
        ka = numpy.zeros( ( len( knots ), len( sl ) ), dtype=float )
        for k,s in enumerate( sl ) :
            ka[:,k] = s.model.knots
        for j in range( len( knots ) ) :
            plt.plot( ka[j,:], cc[j % 6] )
        plt.show()

    def test3( self ) :
        print( "  Test Dynamic Splines" )

        t = numpy.linspace( 0, 100, 201, dtype=float )
        y = numpy.sin( 2 * math.pi * numpy.exp( t / 150 ) )

        knots =[0, 25, 50, 75, 100]

        mdl = SplinesDynamicModel( knots=knots )
        mdl.setLimits( lowLimits=[-2.0], highLimits=[+2.0] )

#        engs = ["galilean", "chord", "struct"]
        engs = ["galilean", "birth", "death", "struct"]

        ns = NestedSampler( t, mdl, y, seed=1234, engines=engs )
        ns.distribution.setLimits( [0.01,100] )
        ns.verbose = 2

        evid = ns.sample( plot=self.doplot )

    def test4( self ) :
        print( "  Test Modifiable Splines compound" )

        NP = 41
        t = numpy.linspace( 0, 100, NP, dtype=float )
        dt = GaussModel().result( t, [10.0, 34, 10] )

        ym = numpy.sin( 0.1 * ( t + dt ) )
        y = ym + 0.1 * numpy.random.randn( NP )

        knots =[0, 25, 50, 75, 100]
        xm = SplinesDynamicModel( knots=knots, dynamic=False )
        xm += PowerModel( 1, fixed={0:1.0} )
        xm.setPrior( 0, LaplacePrior( center=0, scale=4 ) )

        sm = SineModel()
        sm.setLimits( lowLimits=[-2.0], highLimits=[+2.0] )

        mdl = xm | sm

        ns = NestedSampler( t, mdl, y, seed=31234 )
        ns.distribution.setLimits( [0.01,100] )
        ns.verbose = 2
        ns.engines[2].verbose = 5
        ns.engines[2].slow = 5

        evid = ns.sample( plot=self.doplot )

        if not self.doplot : return

        cc = ['k,', 'b,', 'r,', 'g,', 'c,', 'm,']
        sl = ns.samples
        yfit = numpy.zeros( NP, dtype=float )
        for s in sl :
            knts = s.model.knots
            yft = BasicSplinesModel( knts ).result( t, s.parameters[:7] )
            yfit += yft * s.weight
        plt.plot( t, yfit, 'b-' )
        plt.plot( t, ns.yfit, 'g-' )
        plt.plot( t, y, 'k.' )
        plt.show()


    def suite( cls ):
        return unittest.TestCase.suite( TestModifiable.__class__ )


if __name__ == '__main__':
    unittest.main()




import numpy as numpy
from . import Tools
from .Tools import setAttribute as setatt

from .LinearModel import LinearModel

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2020 Do Kester

class PolySurfaceModel( LinearModel ):
    """
    General polynomial surface model of arbitrary degree.

        f( x,y:p ) = &sum;_d &sum;_k p_n * x^{d-k} * y^k )

    where the first sum is over d running from 0 to degree ( inclusive )
    and the second sum is over k running from 0 to d ( inclusive ).
    The index n is just incrementing, making all p's different.

    It is a 2-dimensional linear model.

    Examples
    --------
    poly = PolySurfaceModel( 3 )         # 3rd degree polynomial
    print poly.getNumberOfParameters( )        # 10

    Author      Do Kester

    Attributes
    ----------
    degree : int
        degree of the polynomial

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    """
    def __init__( self, degree, copy=None, **kwargs ):
        """
        Polynominal surface of a certain degree.

        The number of parameters is ( degree+2 ) * ( degree+1 ) / 2

        Parameters
        ----------
        degree : int
            the degree of the polynomial.
        copy : PolySurfaceModel
            model to be copied
        fixed : dictionary of {int:float|Model}
            int             list if parameters to fix permanently. Default None.
            float|Model     list of values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See @FixedModel

        """
        super( PolySurfaceModel, self ).__init__( (degree+2)*(degree+1)//2, ndim=2,
                copy=copy, **kwargs )
        self.degree = degree


    def copy( self ):
        """ Copy method.  """
        return PolySurfaceModel( self.degree, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: degree
        """
        if name == 'degree' :
            setatt( self, name, value, type=int )
        else :
            super( PolySurfaceModel, self ).__setattr__( name, value )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input values.

        The partials are the powers of x,y ( xdata ) from 0 to degree.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        nx = Tools.length( xdata[:,0] )
        part = numpy.zeros( ( nx, self.npmax ), dtype=float )
        n = 0
        for d in range( self.degree + 1 ) :
            x = numpy.ones( nx, dtype=float )
            i = 0
            while i <= d :
                part[:,n+i] = x
                x *= xdata[:,1]
                i += 1
            x = numpy.ones( nx, dtype=float )
            i = d
            while i >= 0 :
                part[:,n+i] *= x
                x *= xdata[:,0]
                i -= 1
            n += d + 1
        if parlist is None :
            return part

        return part[:,parlist]

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        bn = "PolySurface: f( x,y:p ) ="
        n = 0
        plus = ""
        for d in range( self.degree + 1 ) :
            i = d
            k = 0
            while k <= d :
                bn += "%s p_%d"%(plus,n)
                if i > 0 :
                    bn += " * x%s"%( "" if i == 1 else "^%d"%i )
                if k > 0 :
                    bn += " * y%s"%( "" if k == 1 else "^%d"%k )
                n += 1
                k += 1
                i -= 1
            plus = " +"
        return bn

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.
        Parameters  k    parameter number.

        """
        dk = 0
        while k > dk:
            dk += 1
            k -= dk
        return self.yUnit / ( ( self.xUnit[0]**k ) * ( self.xUnit[1] ** (dk-k) ) )






from .UniformPrior import UniformPrior

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 - 2020 Do Kester

class CircularUniformPrior( UniformPrior ):
    """
    Cricular Uniform prior distribution, for location parameters.
    The lowLimit is wrapped onto the highLimit.

    A circular uniform prior is a proper prior ( i.e. its integral is bound ).
    Because of its wrapping around needs limits, low and high, such that
    -Inf < low < high < +Inf.

        Pr( x ) = 1 / ( high - low )   if low < x < high else 0

    For computational purposes the unit range is from [1/3..2/3]; the u value
    always returns in that range. The wings are needed for the wrapping.

    The d value aways returns inside the range [low..high]

    domain2Unit: u = ( d - lo ) / range
                 u = ( u + 1 ) / 3              ## shrink to [1/3..2/3]
    The u value can reach any value in [0..1].
    unit2Domain: u = ( 3 * u ) % 1.0            ## wrap around
                 d = u  * range + lo

    Attributes from UniformPrior
    ----------------------------
    _range : float
        valid range ( highLimit - lowLimit )

    Attributes from Prior
    --------------------=
    lowLimit, highLimit, deltaP, _lowDomain, _highDomain

    """

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, limits=None, prior=None ):
        """
        Constructor.

        Parameters
        ----------
        limits : array of 2 floats
            [low,high]  range of the prior. Low is wrapped onto high.
        prior : CircularUniformPrior
            to be copied.

        """
        super( CircularUniformPrior, self ).__init__( limits=limits, prior=prior )

    def copy( self ):
        """ Return a (deep) copy of itself. """
        return CircularUniformPrior( prior=self, limits=[self.lowLimit,self.highLimit] )

    def isCircular( self ) :
        """
        Always True
        """
        return True

    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for a Uniform distribution.

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        uval = super( CircularUniformPrior, self ).domain2Unit( dval )
        return ( uval + 1 ) / 3.0

    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for a Uniform distribution.

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        uval = ( uval * 3 ) % 1.0
        return super( CircularUniformPrior, self ).unit2Domain( uval )


    def __str__( self ):
        """ Return a string representation of the prior.  """
        return str( "CircularUniformPrior " + ( "unbound." if not self.isBound( )
                else ( "period [%.2f..%.2f]"%( self.lowLimit, self.highLimit ) ) ) )






import numpy as numpy

from . import Tools
from .Formatter import formatter as fmt
from .Dynamic import Dynamic
from .Engine import Engine

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL"
__version__ = "2.6.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2018 - 2020 Do Kester


class DeathEngine( Engine ):
    """
    The DeathEngine deletes a component from the model.

    Only for Models that are Dynamic.
    The death rate is governed by the growth-prior in the dynamic model.

    The member is kept when the logLikelihood > lowLhood.

    Author       Do Kester.

    """
#    _deathrate = 1.0

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, slow=None, copy=None, seed=23455, verbose=0 ):
        """
        Constructor.

        Parameters
        ----------
        walkers : WalkerList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        slow : None or int > 0
            Run this engine every slow-th iteration. None for all.
        copy : GalileanEngine
            to be copied
        seed : int
            for random number generator
        """
        super( ).__init__( walkers, errdis, slow=slow, copy=copy,
                    seed=seed, verbose=verbose )

    def copy( self ):
        """ Return copy of this.  """
        return DeathEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "DeathEngine" )

    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood, append=False ):
        """
        Execute the engine by removins a component.

        Parameters
        ----------
        kw : int
            index of walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        append : bool
            set walker in place or append

        Returns
        -------
        int : the number of successfull moves

        """
        self.reportCall()

        walker = self.walkers[kw].copy()          ## work on local copy

        nhyp = self.errdis.nphypar
        walker.problem.model.parameters = walker.allpars[:-nhyp]


        problem = walker.problem
        model = problem.model
        allp = walker.allpars
        ptry = allp
        ftry = walker.fitIndex

        off = 0
        while model is not None and not isinstance( model, Dynamic ) :
            off += model.npbase
            model = model._next

        nc = model.ncomp
        np = model.npbase

        if self.verbose > 4 :
            print( "DEN1  ", walker.id, nc, np, len( ptry ), len( ftry ) )

        # shuffle the parameters (if needed) before throwing the last one out.
        ptry = model.shuffle( ptry, off, np, self.rng )


        if not ( nc > model.growPrior.unit2Domain( self.rng.rand() ) and
                model.shrink( offset=off, rng=self.rng ) ) :
            self.reportFailed()
            return 0

        dnp = model.npbase - np         # parameter decrease

        ftry = model.alterFitindex( ftry, np, dnp, off )
        ptry = problem.model.parameters
        if nhyp > 0 :
            ptry = numpy.append( ptry, allp[-nhyp:] )

        Ltry = self.errdis.logLikelihood( problem, ptry )

        if Ltry >= lowLhood:
            self.reportSuccess()
            update = len( self.walkers ) if append else kw
            self.setWalker( update, problem, ptry, Ltry, fitIndex=ftry )
            wlkr = self.walkers[update]
            wlkr.check( nhyp=nhyp )
            return abs( dnp )

        self.reportReject( )

        return 0


